import numpy as np
import cvxopt
import cvxopt.solvers


class SVM(object):

    def __init__(self, C=1.0):
        self.C = float(C)

    def fit(self, X, y):
        n_samples, n_features = X.shape

        Xy = X * y[:, np.newaxis]
        P = cvxopt.matrix(np.dot(Xy, Xy.T))
        q = cvxopt.matrix(np.ones(n_samples) * -1)
        A = cvxopt.matrix(y, (1, n_samples))
        b = cvxopt.matrix(0.0)

        tmp1 = np.diag(np.ones(n_samples) * -1)
        tmp2 = np.identity(n_samples)
        G = cvxopt.matrix(np.vstack((tmp1, tmp2)))
        tmp1 = np.zeros(n_samples)
        tmp2 = np.ones(n_samples) * self.C
        h = cvxopt.matrix(np.hstack((tmp1, tmp2)))

        # solve QP model
        solution = cvxopt.solvers.qp(P, q, G, h, A, b)

        # Lagrange multipliers
        a = np.ravel(solution['x'])

        # Support vectors have non zero lagrange multipliers
        sv = a > 1e-5
        self.a = a[sv]
        self.sv = X[sv]
        self.sv_y = y[sv]
        print(("%d support vectors out of %d points" % (len(self.a), n_samples)))

        # Intercept
        self.b = 0
        for n in range(len(self.a)):
            self.b += self.sv_y[n]
            self.b -= np.sum(self.a * self.sv_y * np.dot(self.sv, self.sv[n]))
        self.b /= len(self.a)

        self.w = np.zeros(n_features)
        for n in range(len(self.a)):
            self.w += self.a[n] * self.sv_y[n] * self.sv[n]

    def decision_function(self, X):
        return np.dot(X, self.w) + self.b

    def predict(self, X):
        return np.sign(self.decision_function(X))
